// asmppc.c

#define versionName "PowerPC assembler"
#include "asmx.h"

enum
{
    CPU_POWER,
    CPU_601,
    CPU_603R
};

enum
{
    o_Branch,
    o_BranchCond,
    o_Jump,
    o_JumpCond,
    o_BCxRx,
    o_OneReg,
    o_OneRegShift,  // Rd,shifter
    o_OneRegShiftS, // Rd,shifter with optional S
    o_TwoReg,
    o_TwoRegShifted,
    o_TwoRegSwap,
    o_ThreeReg,
    o_ThreeRegSwap,
    o_BKPT,         // BKPT immed16
    o_BLX,
    o_CMP,
    o_CMPI,
    o_CLZ,
    o_CR,
    o_MUL,          // MUL
    o_SWP,          // SWP
    o_LDM_STM,      // LDM/STM
    o_LDR_STR,      // LDR/STR
    o_MRS,          // MRS
    o_MSR,          // MSR
    o_MLA,          // MLA
    o_MLAL,         // SMLAL/SMULL/UMLAL/UMULL
    o_CDP,          // CDP
    o_MCR_MRC,      // MCR/MRC
    o_LDC_STC,      // LDC/STC
    o_ADR,          // ADR/ADRL
    o_Imm,
    o_ImmSwap,
    o_Implied,      // NOP

//  o_Foo = o_LabelOp,
};

const char PPC_regs[] = "R0 R1 R2 R3 R4 R5 R6 R7 R8 R9 R10 R11 R12 R13 R14 R15 R16 R17 R18 R19 R20 R21 R22 R23 R24 R25 R26 R27 R28 R29 R30 R31";
const char fregs[] = "F0 F1 F2 F3 F4 F5 F6 F7 F8 F9 F10 F11 F12 F13 F14 F15 F16 F17 F18 F19 F20 F21 F22 F23 F24 F25 F26 F27 F28 F29 F30 F31";
const char sregs[] = "SR0 SR1 SR2 SR3 SR4 SR5 SR6 SR7 SR8 SR9 SR10 SR11 SR12 SR13 SR14 SR15";
const char PPCflags[] = "O .";
const char other[] = "PC";

struct OpcdRec POWER_opcdTab[] =
{
    {"BCC",     o_BCxRx,        0x4C000420},
    {"BCR",     o_BCxRx,        0x4C000020},
    
    {"AME",     o_TwoReg,       0x7C0001D4},
    {"AZE",     o_TwoReg,       0x7C000194},
    
    {"A",       o_ThreeReg,     0x7C000014},
    {"AE",      o_ThreeReg,     0x7C000114},
    {"CAX",     o_ThreeReg,     0x7C000214},
    
    {"AI",      o_Imm,          0x30000000},
    {"CAL",     o_Imm,          0x38000000},
    {"CAU",     o_Imm,          0x3C000000},
    
    {"ANDIL.",  o_ImmSwap,      0x70000000},
    {"ANDIU.",  o_ImmSwap,      0x74000000},
};

struct OpcdRec PPC_opcdTab[] =
{
    // B and BL don't use wildcards to avoid false matches with BL, BLX, and BX
    {"B",       o_Branch,       0x48000000}, // rewrited
    {"BC",      o_BranchCond,   0x40000000}, // rewrited
    
    {"BCCTR",   o_BCxRx,        0x4C000420}, // rewrited
    {"BCLR",    o_BCxRx,        0x4C000020}, // rewrited
    
    {"CMP",     o_CMP,          0x7C000000}, // rewrited
    {"CMPL",    o_CMP,          0x7C000040}, // rewrited
    {"CMPI",    o_CMPI,         0x2C000000}, // rewrited
    {"CMPLI",   o_CMPI,         0x28000000}, // rewrited
    
    {"CRAND",   o_CR,           0x4C000202}, // rewrited
    {"CRANDC",  o_CR,           0x4C000102}, // rewrited
    {"CREQV",   o_CR,           0x4C000242}, // rewrited
    {"CRNAND",  o_CR,           0x4C0001C2}, // rewrited
    {"CRNOR",   o_CR,           0x4C000042}, // rewrited
    {"CROR",    o_CR,           0x4C000382}, // rewrited
    {"CRORC",   o_CR,           0x4C000342}, // rewrited
    {"CRXOR",   o_CR,           0x4C000182}, // rewrited

    //{"BX*",     o_OneReg,       0x012FFF10}, // ARM5/ARM4T

    {"CMN*",    o_OneRegShift,   0x01700000},
    {"TEQ*",    o_OneRegShift,   0x01300000},
    {"TST*",    o_OneRegShift,   0x01100000},

    {"MOV*",    o_OneRegShiftS,  0x01A00000},
    {"MVN*",    o_OneRegShiftS,  0x01E00000},

    {"DCBF",    o_TwoRegShifted, 0x7C0000AC}, // rewrited
    {"DCBI",    o_TwoRegShifted, 0x7C0003AC}, // rewrited
    {"DCBST",   o_TwoRegShifted, 0x7C00006C}, // rewrited
    {"DCBT",    o_TwoRegShifted, 0x7C00022C}, // rewrited
    {"DCBTST",  o_TwoRegShifted, 0x7C0001EC}, // rewrited
    {"DCBZ",    o_TwoRegShifted, 0x7C0007EC}, // rewrited
    
    {"ABS",     o_TwoReg,       0x7C0002D0}, // rewrited
    {"ADDME",   o_TwoReg,       0x7C0001D4}, // rewrited
    {"ADDZE",   o_TwoReg,       0x7C000194}, // rewrited
    {"CLCS",    o_TwoReg,       0x7C000426}, // rewrited
    
    {"CNTLZW",  o_TwoRegSwap,   0x7C000034}, // rewrited
    
    {"ADD",     o_ThreeReg,     0x7C000214}, // rewrited
    {"ADDC",    o_ThreeReg,     0x7C000014}, // rewrited
    {"ADDE",    o_ThreeReg,     0x7C000114}, // rewrited
    {"DIV",     o_ThreeReg,     0x7C000296}, // rewrited
    {"DIVS",    o_ThreeReg,     0x7C0002D6}, // rewrited
    
    {"AND",     o_ThreeRegSwap, 0x7C000038}, // rewrited
    {"ANDC",    o_ThreeRegSwap, 0x7C000078}, // rewrited

    {"BKPT",    o_BKPT,         0xE1200070},
    
    {"ADDI",    o_Imm,          0x38000000}, // rewrited
    {"ADDIC",   o_Imm,          0x30000000}, // rewrited
    {"ADDIS",   o_Imm,          0x3C000000}, // rewrited
    
    {"ANDI.",   o_ImmSwap,      0x70000000}, // rewrited
    {"ANDIS.",  o_ImmSwap,      0x74000000}, // rewrited

    {"BLX*",    o_BLX,          0x012FFF30}, // ARM5

    {"CLZ*",    o_CLZ,          0x016F0F10}, // ARM5

    {"MUL*",    o_MUL,          0x00000090}, // ARM2

    {"SWP*",    o_SWP,          0x01000090}, // ARM2a/ARM3

    {"LDM*",    o_LDM_STM,      0x00100000},
    {"STM*",    o_LDM_STM,      0x00000000},

    {"LDR*",    o_LDR_STR,      0x00100000}, // LDRH/LDRSB/LDRSH = ARM4
    {"STR*",    o_LDR_STR,      0x00000000}, // STRH = ARM4

    {"MRS*",    o_MRS,          0x010F0000}, // ARM3

    {"MSR*",    o_MSR,          0x0120F000}, // ARM3

    {"MLA*",    o_MLA,          0x00200090}, // ARM2

    {"SMLAL*",  o_MLAL,         0x00E00090}, // ARM-M
    {"SMULL*",  o_MLAL,         0x00C00090}, // ARM-M
    {"UMLAL*",  o_MLAL,         0x00A00090}, // ARM-M
    {"UMULL*",  o_MLAL,         0x00800090}, // ARM-M

    {"CDP*",    o_CDP,          0x0E000000}, // ARM2/5

    {"MCR*",    o_MCR_MRC,      0x0E000010}, // ARM2/5
    {"MRC*",    o_MCR_MRC,      0x0E100010}, // ARM2/5

    {"LDC*",    o_LDC_STC,      0x0C100000}, // ARM2/5
    {"STC*",    o_LDC_STC,      0x0C000000}, // ARM2/5

    {"ADR*",    o_ADR,          0},          // pseudo-instruction

    {"NOP",     o_Implied,      0xE1A00000}, // pseudo-instruction = MOV R0,R0

    {"",        o_Illegal,      0}
};


// --------------------------------------------------------------


long PPCImmed(unsigned long val)
{
    // note: can't abort assembling instruction because it may cause phase errors

    if ((val & 0xFFFFFF00) == 0)
        return (val & 0xFF);
    else
    {
        // FIXME: need to handle shifts
        // bit 25 (I) = 1
        // bits 8-11 = rotate (immed8 is rotated right by 2 * this value)
        // bits 0-7 = immed8
        // the smallest value of rotate should be chosen
    }

    Error("Invalid immediate constant");
    return 0;
}


bool PPCShifter(unsigned long *shift)
{
    int     reg1,reg2,typ;
    Str255  word;
    char    *oldLine;
    int     token;
    int     val;

    *shift = 0;

    token = GetWord(word);
    
    reg1 = FindReg(word,PPC_regs);
    if (reg1 > 15) reg1 = reg1 - 3; // SP LR PC -> R13 R14 R15
    if (CheckReg(reg1)) return 1;

    oldLine = linePtr;
    token = GetWord(word);
    if (token != ',')
    {
    linePtr = oldLine;
    *shift = reg1;
    }
    else
    {
     // shifts and rotates
     //typ = GetReg(shifts);
     if (typ == 5) typ = 0; // ASL -> LSL
     if (typ < 0)
        IllegalOperand();
     else if (typ == 4)
     {
      *shift = 0x60 | reg1; // RRX
     }
     else
     {
      token = GetWord(word);
      if (token == '#')
      {
       val = Eval();
       // FIXME: need to range check immed5
       *shift = (typ << 5) | reg1 | ((val & 0x1F) << 7);
      }
      else
      {
       reg2 = FindReg(word,PPC_regs);
       if (reg2 > 15) reg2 = reg2 - 3; // SP LR PC -> R13 R14 R15
       if (CheckReg(reg2)) return 1;

       *shift = (typ << 5) | (1 << 4) | reg1 | (reg2 << 8);
       }
       }
       }

    return 0;
}


int PPCEvalBranch(int width) // instrLen should be 8
{
    long val;
    long limit;

    limit = (1 << (width+1)) - 1;

    val = Eval();
    val = val - locPtr;
    if (!errFlag && ((val & 3) || val < ~limit || val > limit))
        Error("Long branch out of range");

    return val;
}


int PPC_DoCPUOpcode(int typ, int parm)
{
    int     val1, val2, val3;
    long    val;
    int     i;
    Str255  word;
    char    *oldLine;
    int     token;
    int     reg1,reg2,reg3,reg4;
    int     cond;
    unsigned long  mode;
    unsigned short regbits;

    // get condition from opcode wildcard
    word[0]=0;
    if (isalnum(*linePtr))
        GetWord(word);

    switch(typ)
    {
        case o_Branch:
            oldLine=linePtr;
            if (GetWord(word) == 'L') parm=parm|1;
            else linePtr=oldLine;
            if (GetWord(word) == 'A') parm=parm|2;

            if(parm&2==0) val = PPCEvalBranch(26);
            else val = Eval();
            InstrL(parm | (val & 0x03FFFFFC));
            break;
            
        case o_BranchCond:
            oldLine=linePtr;
            if (GetWord(word) == 'L') parm=parm|1;
            else linePtr=oldLine;
            if (GetWord(word) == 'A') parm=parm|2;
            
            val1 = Eval();
            
            if (Comma()) break;
            
            val2 = Eval();
            
            if (Comma()) break;
            
            if(parm&2==0)
            {
             val3 = PPCEvalBranch(16);
             InstrL(parm | ((val1 & 31) << 21) | ((val2 & 31) << 16) | (val3 & 0x0FFFC));
            }
            else
            {
             val = Eval();
             InstrL(parm | ((val1 & 31) << 21) | ((val2 & 31) << 16) | (val & 0x0FFFC));
            }
            break;

        case o_BCxRx:
            if (GetWord(word) == 'L') parm=parm|1;
            
            val1 = Eval();
            
            if (Comma()) break;
            
            val2 = Eval();

            InstrL(parm | ((val1 & 31) << 21) | ((val2 & 31) << 16));
            break;

        case o_OneReg:       // Rm
            reg1 = GetReg(PPC_regs);
            if (reg1 > 15) reg1 = reg1 - 3; // SP LR PC -> R13 R14 R15
            if (CheckReg(reg1)) break;

            InstrL(parm | (cond << 28) | reg1);
            break;

        case o_OneRegShiftS: // Rd,shifter with optional S
            reg1 = GetReg(PPC_regs);
            if (reg1 > 15) reg1 = reg1 - 3; // SP LR PC -> R13 R14 R15
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            if (PPCShifter(&mode)) break;

            InstrL(parm | (cond << 28) | (val1 << 20) | (reg1 << 12) | mode);
            break;

        case o_Imm:
            if(parm==0x30000000)
            {
             if (GetWord(word) == '.') parm=parm|0x4000000;
            }
            
            reg1 = GetReg(PPC_regs);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(PPC_regs);
            if (CheckReg(reg2)) break;

            if (Comma()) break;

            val1 = Eval();

            InstrL(parm | (reg1 << 21) | (reg2 << 16) | val1);
            break;

        case o_ImmSwap:
            reg1 = GetReg(PPC_regs);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(PPC_regs);
            if (CheckReg(reg2)) break;

            if (Comma()) break;

            val1 = Eval();

            InstrL(parm | (reg2 << 21) | (reg1 << 16) | val1);
            break;

        case o_TwoReg:
            if((parm==0x7C0002D0||parm==0x7C000426)&&curCPU>CPU_601) Error("This instruction is not used after PowerPC 601");

            if(parm!=0x7C0002D0&&parm!=0x7C000426)
            {
             oldLine=linePtr;
             if (GetWord(word) == 'O') parm=parm|0x400;
             else linePtr=oldLine;
             if (GetWord(word) == '.') parm=parm|1;
            }

            reg1 = GetReg(PPC_regs);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(PPC_regs);
            if (CheckReg(reg2)) break;

            InstrL(parm | (reg1 << 21) | (reg2 << 16));
            break;

        case o_TwoRegShifted:
            reg1 = GetReg(PPC_regs);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(PPC_regs);
            if (CheckReg(reg2)) break;

            InstrL(parm | (reg1 << 16) | (reg2 << 11));
            break;

        case o_TwoRegSwap:
            if (GetWord(word) == '.') parm=parm|1;
            
            reg1 = GetReg(PPC_regs);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(PPC_regs);
            if (CheckReg(reg2)) break;

            InstrL(parm | (reg2 << 21) | (reg1 << 16));
            break;

        case o_ThreeReg:
            oldLine=linePtr;
            GetWord(word);
            reg1=FindReg(word,"O");
            if(reg1==0) parm=parm|0x400;
            else linePtr=oldLine;
            
            GetWord(word);
            reg1=FindReg(word,".");
            if(reg1==0) parm=parm|1;
            else linePtr=oldLine;
            
            reg1 = GetReg(PPC_regs);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(PPC_regs);
            if (CheckReg(reg2)) break;

            if (Comma()) break;

            reg3 = GetReg(PPC_regs);
            if (CheckReg(reg3)) break;

            InstrL(parm | (reg1 << 21) | (reg2 << 16) | (reg3 << 11));
            break;

        case o_ThreeRegSwap:
            if (GetWord(word) == '.') parm=parm|1;
             
            reg1 = GetReg(PPC_regs);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(PPC_regs);
            if (CheckReg(reg2)) break;

            if (Comma()) break;

            reg3 = GetReg(PPC_regs);
            if (CheckReg(reg3)) break;

            InstrL(parm | (reg2 << 21) | (reg1 << 16) | (reg3 << 11));
            break;

        case o_BKPT:         // BKPT immed16
            val1 = Eval();
            // FIXME: need to check immed16 range

            InstrL(parm  | ((val1 << 4) & 0x000FFF00) | (val1 & 0x0F));
            break;

        case o_BLX:          // BLX
            //if (ARMCond(&cond,word)) return 0;

            oldLine = linePtr;

            reg1 = GetReg(PPC_regs);
            if (reg1 > 15) reg1 = reg1 - 3; // SP LR PC -> R13 R14 R15

            if (reg1 >= 0)
            {
                InstrL(parm | (cond << 28) | reg1);
            }
            else
            {
                linePtr = oldLine;

                if (cond != 14) return 0; // this form of BLX is always unconditional

                val1 = PPCEvalBranch(25);
                InstrL(0xFA000000 | ((val1 << 23) & 0x01000000) | ((val1 >> 2) & 0x00FFFFFF));
            }

            break;

        case o_CMP:
            val1 = Eval();
            
            if (Comma()) break;

            val2 = Eval();
            
            if (Comma()) break;

            reg1 = GetReg(PPC_regs);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(PPC_regs);
            if (CheckReg(reg2)) break;

            InstrL(parm | ((val1 & 7) << 23) | ((val2 & 1) << 21) | (reg1 << 16) | (reg2 << 11));
            break;

        case o_CMPI:
            val1 = Eval();
            
            if (Comma()) break;

            val2 = Eval();
            
            if (Comma()) break;

            reg1 = GetReg(PPC_regs);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            val3 = Eval();

            InstrL(parm | ((val1 & 7) << 23) | ((val2 & 1) << 21) | (reg1 << 16) | val3);
            break;

        case o_CLZ:          // CLZ
            //if (ARMCond(&cond,word)) return 0;

            reg1 = GetReg(PPC_regs);
            if (reg1 > 15) reg1 = reg1 - 3; // SP LR PC -> R13 R14 R15
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(PPC_regs);
            if (reg2 > 15) reg2 = reg2 - 3; // SP LR PC -> R13 R14 R15
            if (CheckReg(reg2)) break;

            InstrL(parm | (cond << 28) | (reg1 << 12) | reg2);
            break;
            
        case o_CR:
            val1 = Eval();
            
            if (Comma()) break;
            
            val2 = Eval();
            
            if (Comma()) break;
            
            val3 = Eval();
            InstrL(parm | ((val1 & 31) << 21) | ((val2 & 31) << 16) | ((val3 & 31)<<11));
            break;

        case o_MUL:          // MUL
            //if (ARMCond(&cond,word)) return 0;

            reg1 = GetReg(PPC_regs);
            if (reg1 > 15) reg1 = reg1 - 3; // SP LR PC -> R13 R14 R15
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(PPC_regs);
            if (reg2 > 15) reg2 = reg2 - 3; // SP LR PC -> R13 R14 R15
            if (CheckReg(reg2)) break;

            if (Comma()) break;

            reg3 = GetReg(PPC_regs);
            if (reg3 > 15) reg3 = reg3 - 3; // SP LR PC -> R13 R14 R15
            if (CheckReg(reg3)) break;

            InstrL(parm | (cond << 28) | (val1 << 20) | (reg1 << 16) | (reg3 << 8) | reg2);
            break;

        case o_MRS:          // MRS
            //if (ARMCond(&cond,word)) return 0;

            reg1 = GetReg(PPC_regs);
            if (reg1 > 15) reg1 = reg1 - 3; // SP LR PC -> R13 R14 R15
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg("CPSR SPSR");
            if (CheckReg(reg2)) break;

            InstrL(parm | (cond << 28) | (reg2 << 22) | (reg1 << 12));
            break;

        case o_ADR:          // ADR/ADRL
// ADRL Rn,address
            if (word[0] == 'L')
            {
                parm = 1;
                //if (ARMCond(&cond,word+1)) return 0;
            }
            else
            {
                //if (ARMCond(&cond,word)) return 0;
            }

            reg1 = GetReg(PPC_regs);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            val1 = EvalLBranch(8);

            i = parm + 1; // number of longwords

            if (evalKnown && parm)
            {
                // *** determine auto size for ADRL
                if (-0xFF <= val1 && val1 <= 0xFF)
                    i = 1;
            }

            switch(i)
            {
                case 4:
                    break;

                case 3:
                    break;

                case 2:
                    if (val1 < 0)
                        InstrLL(0x028F0000 | (cond << 28) | (reg1 << 12) | (-val1 & 0xFF),
                                0x02800400 | (cond << 28) | (reg1 << 16) | (reg1 << 12) | ((-val1 >> 8) & 0xFF));
                    else
                        InstrLL(0x024F0000 | (cond << 28) | (reg1 << 12) | (val1 & 0xFF),
                                0x02400000 | (cond << 28) | (reg1 << 16) | (reg1 << 12) | ((val1 >> 8) & 0xFF));
                    break;

                default:
                    if (val1 < 0)
                        InstrL(0x028F0000 | (cond << 28) | (reg1 << 12) | (-val1 & 0xFF));
                    else
                        InstrL(0x024F0000 | (cond << 28) | (reg1 << 12) | (val1 & 0xFF));
                    break;
            }

// gets complicated because of offsets

//   ADR     R1,.            ; E24F1008
//   ADRL    R1,.            ; E24F1008
//   ADRL    R1,.+$1000      ; E28F1FFE E2811B03
//   ADRL    R1,.+$100000    ; E28F1FFE E2811BFF E2811703
//   ADRL    R1,.+$10000000  ; E28F1FFE E2811BFF E28117FF E2811303

if (pass == 2)
{
    printf("*** %.8X *** %s\n",val1,line);
}

//            InstrL(   (cond << 28) | (val1 & 0x0FFF));
            break;

        case o_Implied:      // NOP
            InstrL(parm);
            break;

        default:
            return 0;
            break;
    }

    if (locPtr & 3)
    {
        Error("Code at non-longword-aligned address");
        // deposit extra bytes to reset alignment and prevent further errors
        for (i=locPtr & 3; i<4; i++)
            InstrAddB(0);
        // note: Inserting bytes in front won't work because offsets have already been assembled.
        // The line could be re-assembled by recursively calling DoCPUOpcode, but then
        // the label before the instruction would still be at the odd address.
    }

    return 1;
}


void AsmPPCInit(void)
{
    char *p;

    p = AddAsm(versionName, &PPC_DoCPUOpcode, NULL, NULL);
    AddCPU(p, "POWER",   0, BIG_END, ADDR_24, LIST_24, 8, 0, POWER_opcdTab);
    AddCPU(p, "PPC601",  0, BIG_END, ADDR_24, LIST_24, 8, 0, PPC_opcdTab);
    AddCPU(p, "PPC603R", 0, BIG_END, ADDR_24, LIST_24, 8, 0, PPC_opcdTab);
}
